// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file StdMpi.h */

#ifndef AMDIS_STDMPI_H
#define AMDIS_STDMPI_H

#include <map>
#include <stdint.h>
#include <mpi.h>
#include "MeshStructure.h"
#include "parallel/InteriorBoundary.h"

namespace AMDiS {

  int intSizeOf(int &data);

  int intSizeOf(std::map<const DegreeOfFreedom*, DegreeOfFreedom> &data);

  int intSizeOf(std::map<WorldVector<double>, int> &data);

  int intSizeOf(std::vector<MeshStructure> &data);

  int intSizeOf(std::vector<WorldVector<double> > &data);

  int intSizeOf(std::vector<int> &data);

  int intSizeOf(std::vector<double> &data);
  
  int intSizeOf(std::vector<std::pair<int, int> > &data);

  int intSizeOf(std::vector<AtomicBoundary> &data);

  int intSizeOf(std::vector<BoundaryObject> &data);

  int intSizeOf(std::vector<const DegreeOfFreedom*> &data);

  int intSizeOf(std::vector<std::vector<double> > &data);

  int intSizeOf(std::map<BoundaryType, std::map<DegreeOfFreedom, DegreeOfFreedom> > &data);

  void makeBuf(int &data, int *buf);

  void makeFromBuf(int &data, int *buf, int bufSize);

  void makeBuf(std::map<const DegreeOfFreedom*, DegreeOfFreedom> &data, int* buf);

  void makeFromBuf(std::vector<std::pair<DegreeOfFreedom, DegreeOfFreedom> > &data, 
		   int *buf, int bufSize);

  void makeBuf(std::map<WorldVector<double>, int> &data, double* buf);

  void makeFromBuf(std::map<WorldVector<double>, int> &data, double* buf, int bufSize);

  void makeBuf(std::vector<MeshStructure> &data, uint64_t *buf);

  void makeFromBuf(std::vector<MeshStructure> &data, 
		   uint64_t *buf, int bufSize);

  void makeBuf(std::vector<WorldVector<double> > &data, double *buf);

  void makeFromBuf(std::vector<WorldVector<double> > &data, double *buf, int bufSize);

  void makeBuf(std::vector<int> &data, int *buf);

  void makeFromBuf(std::vector<int> &data, int *buf, int bufSize);

  void makeBuf(std::vector<double> &data, double *buf);

  void makeFromBuf(std::vector<double> &data, double *buf, int bufSize);

  void makeBuf(std::vector<std::pair<int, int> > &data, int *buf);

  void makeBuf(std::vector<AtomicBoundary> &data, int *buf);

  void makeFromBuf(std::vector<AtomicBoundary> &data, int *buf, int bufSize);

  void makeBuf(std::vector<BoundaryObject> &data, int *buf);

  void makeFromBuf(std::vector<BoundaryObject> &data, int *buf, int bufSize);

  void makeBuf(std::vector<std::vector<double> > &data, double *buf);

  void makeFromBuf(std::vector<std::vector<double> > &data, double *buf, int bufSize);
  
  void makeBuf(std::map<BoundaryType, std::map<DegreeOfFreedom, DegreeOfFreedom> > &data, int *buf);

  void makeFromBuf(std::map<BoundaryType, std::map<DegreeOfFreedom, DegreeOfFreedom> > &data, int *buf, int bufSize);


  template<typename SendT, typename RecvT=SendT>
  class StdMpi 
  {
  public:
    StdMpi(MPI::Intracomm &comm, bool b = true) :
      mpiComm(comm),
      commPrepared(true),
      exchangeDataSize(b)
    {}

    void clear()
    {
      sendData.clear();
      recvData.clear();
      sendDataSize.clear();
      recvDataSize.clear();

      commPrepared = true;
    }

    void send(int toRank, SendT& data)
    {
      FUNCNAME("StdMpi::send()");
      
      TEST_EXIT_DBG(commPrepared)("Communication is not prepared!\n");

      sendData[toRank] = data;
      sendDataSize[toRank] = intSizeOf(data);
    }

    void send(std::map<int, SendT>& data)
    {
      FUNCNAME("StdMpi::send()");
      
      TEST_EXIT_DBG(commPrepared)("Communication is not prepared!\n");

      for (typename std::map<int, SendT>::iterator it = data.begin(); 
	   it != data.end(); ++it) {
	sendData[it->first] = it->second;
	sendDataSize[it->first] = intSizeOf(it->second);
      }
    }

    std::map<int, SendT>& getSendData()
    {
      return sendData;
    }

    SendT& getSendData(int rank)
    {
      return sendData[rank];
    }

    void updateSendDataSize()
    {
      for (typename std::map<int, SendT>::iterator it = sendData.begin(); 
	   it != sendData.end(); ++it)
	sendDataSize[it->first] = intSizeOf(it->second);
    }

    void recv(int fromRank, int size = -1)
    {
      FUNCNAME("StdMpi::recv()");
      
      TEST_EXIT_DBG(commPrepared)("Communication is not prepared!\n");

      recvDataSize[fromRank] = size;
    }

    void recvFromAll(int size = -1)
    {
      FUNCNAME("StdMpi::recvFromAll()");
      
      TEST_EXIT_DBG(commPrepared)("Communication is no prepared!\n");

      for (int i = 0; i < mpiComm.Get_size(); i++)
	recvDataSize[i] = size;
    }

    template<class T>
    void recv(std::map<int, T> &fromRanks)
    {
      FUNCNAME("StdMpi::recv()");
      
      TEST_EXIT_DBG(commPrepared)("Communication is not prepared!\n");

      for (typename std::map<int, T>::iterator it = fromRanks.begin();
	   it != fromRanks.end(); ++it)
	recvDataSize[it->first] = (exchangeDataSize ? -1 : intSizeOf(it->second));
    }

    std::map<int, RecvT>& getRecvData()
    {
      return recvData;
    }

    RecvT& getRecvData(int rank)
    {
      return recvData[rank];
    }

    void commDataSize()
    {
      MPI::Request request[sendData.size() + recvDataSize.size()];
      std::vector<int> sendBuffers, recvBuffers;
      int requestCounter = 0;

      int i = 0;
      for (typename std::map<int, int>::iterator sendIt = sendDataSize.begin();
	   sendIt != sendDataSize.end(); ++sendIt) {
	sendBuffers.push_back(sendIt->second);
	request[requestCounter++] =
	  mpiComm.Isend(&(sendBuffers[i]), 1, MPI_INT, sendIt->first, 0);
	i++;
      }

      for (std::map<int, int>::iterator recvIt = recvDataSize.begin();
	   recvIt != recvDataSize.end(); ++recvIt)
	request[requestCounter++] =
	  mpiComm.Irecv(&(recvIt->second), 1, MPI_INT, recvIt->first, 0);
      
      MPI::Request::Waitall(requestCounter, request);      
    }

    template<class T>
    void startCommunication(MPI_Datatype m)
    {
      FUNCNAME("StdMpi::startCommunication()");

      TEST_EXIT_DBG(commPrepared)("Communication is not prepared!\n");

      if (exchangeDataSize)
	commDataSize();

      MPI::Request request[sendData.size() + recvDataSize.size()];
      int requestCounter = 0;
      std::vector<T*> sendBuffers, recvBuffers;

      for (typename std::map<int, SendT>::iterator sendIt = sendData.begin();
	   sendIt != sendData.end(); ++sendIt) {
	int bufferSize = sendDataSize[sendIt->first];
	T* buf = new T[bufferSize];
	makeBuf(sendIt->second, buf);

  	request[requestCounter++] = 
  	  mpiComm.Isend(buf, bufferSize, m, sendIt->first, 0);

	sendBuffers.push_back(buf);
      }

      for (std::map<int, int>::iterator recvIt = recvDataSize.begin();
	   recvIt != recvDataSize.end(); ++recvIt) {
	T* buf = new T[recvIt->second];

  	request[requestCounter++] =
  	  mpiComm.Irecv(buf, recvIt->second, m, recvIt->first, 0);
	
	recvBuffers.push_back(buf);	
      }

      MPI::Request::Waitall(requestCounter, request);

      for (int i = 0; i < static_cast<int>(sendBuffers.size()); i++)
	delete [] sendBuffers[i];
      sendBuffers.clear();

      int i = 0;
      for (std::map<int, int>::iterator recvIt = recvDataSize.begin();
	   recvIt != recvDataSize.end(); ++recvIt) {

	makeFromBuf(recvData[recvIt->first], recvBuffers[i], recvIt->second);
	delete [] recvBuffers[i];
	i++;
      }      

      commPrepared = false;
    }

  protected:
    ///
    MPI::Intracomm mpiComm;

    ///
    std::map<int, SendT> sendData;

    ///
    std::map<int, RecvT> recvData;

    std::map<int, int> sendDataSize;

    std::map<int, int> recvDataSize;

    bool commPrepared;

    bool exchangeDataSize;
  };

}

#endif
